@@ -59,15 +59,10 @@ fn get_device(ctx: &mut args::ExecCtx<'_>) -> Result<nitrokey::DeviceWrapper> {
   set_log_level(ctx);
 
   match ctx.model {
-    Some(model) => match model {
-      args::DeviceModel::Pro => nitrokey::Pro::connect().map(nitrokey::DeviceWrapper::Pro),
-      args::DeviceModel::Storage => {
-        nitrokey::Storage::connect().map(nitrokey::DeviceWrapper::Storage)
-      }
-    },
+    Some(model) => nitrokey::connect_model(model.into()),
     None => nitrokey::connect(),
   }
-  .map_err(|_| Error::Error("Nitrokey device not found".to_string()))
+  .map_err(|_| Error::from("Nitrokey device not found"))
 }
 
 /// Connect to a Nitrokey Storage device and return it.
@@ -76,27 +71,28 @@ fn get_storage_device(ctx: &mut args::ExecCtx<'_>) -> Result<nitrokey::Storage>
 
   if let Some(model) = ctx.model {
     if model != args::DeviceModel::Storage {
-      return Err(Error::Error(
-        "This command is only available on the Nitrokey Storage".to_string(),
+      return Err(Error::from(
+        "This command is only available on the Nitrokey Storage",
       ));
     }
   }
 
-  nitrokey::Storage::connect().or_else(|_| {
-    Err(Error::Error(
-      "Nitrokey Storage device not found".to_string(),
-    ))
-  })
+  nitrokey::Storage::connect().or_else(|_| Err(Error::from("Nitrokey Storage device not found")))
 }
 
 /// Open the password safe on the given device.
-fn get_password_safe<'dev>(
+fn get_password_safe<'dev, D>(
   ctx: &mut args::ExecCtx<'_>,
-  device: &'dev dyn Device,
-) -> Result<nitrokey::PasswordSafe<'dev>> {
+  device: &'dev D,
+) -> Result<nitrokey::PasswordSafe<'dev>>
+where
+  D: Device,
+{
+  let pin_entry = pinentry::PinEntry::from(pinentry::PinType::User, device)?;
+
   try_with_pin_and_data(
     ctx,
-    pinentry::PinType::User,
+    &pin_entry,
     "Could not access the password safe",
     (),
     |_, pin| device.get_password_safe(pin).map_err(|err| ((), err)),
@@ -117,7 +113,9 @@ where
   D: Device,
   F: Fn(D, &str) -> result::Result<A, (D, nitrokey::CommandError)>,
 {
-  try_with_pin_and_data(ctx, pin_type, msg, device, op)
+  let pin_entry = pinentry::PinEntry::from(pin_type, &device)?;
+
+  try_with_pin_and_data(ctx, &pin_entry, msg, device, op)
 }
 
 /// Authenticate the given device with the user PIN.
@@ -176,7 +174,7 @@ fn get_volume_status(status: &nitrokey::VolumeStatus) -> &'static str {
 /// second or third try, it will call `op` with the data returned by the
 /// previous call to `op`.
 fn try_with_pin_and_data_with_pinentry<D, F, R>(
-  pin_type: pinentry::PinType,
+  pin_entry: &pinentry::PinEntry,
   msg: &'static str,
   data: D,
   op: F,
@@ -188,12 +186,12 @@ where
   let mut retry = 3;
   let mut error_msg = None;
   loop {
-    let pin = pinentry::inquire_pin(pin_type, pinentry::Mode::Query, error_msg)?;
+    let pin = pinentry::inquire(pin_entry, pinentry::Mode::Query, error_msg)?;
     match op(data, &pin) {
       Ok(result) => return Ok(result),
       Err((new_data, err)) => match err {
         nitrokey::CommandError::WrongPassword => {
-          pinentry::clear_pin(pin_type)?;
+          pinentry::clear(pin_entry)?;
           retry -= 1;
 
           if retry > 0 {
@@ -212,7 +210,7 @@ where
 /// Try to execute the given function with a PIN.
 fn try_with_pin_and_data<D, F, R>(
   ctx: &mut args::ExecCtx<'_>,
-  pin_type: pinentry::PinType,
+  pin_entry: &pinentry::PinEntry,
   msg: &'static str,
   data: D,
   op: F,
@@ -220,7 +218,7 @@ fn try_with_pin_and_data<D, F, R>(
 where
   F: Fn(D, &str) -> result::Result<R, (D, nitrokey::CommandError)>,
 {
-  let pin = match pin_type {
+  let pin = match pin_entry.pin_type() {
     pinentry::PinType::Admin => &ctx.admin_pin,
     pinentry::PinType::User => &ctx.user_pin,
   };
@@ -234,7 +232,7 @@ where
     })?;
     op(data, &pin).map_err(|(_, err)| get_error(msg, err))
   } else {
-    try_with_pin_and_data_with_pinentry(pin_type, msg, data, op)
+    try_with_pin_and_data_with_pinentry(pin_entry, msg, data, op)
   }
 }
 
@@ -244,14 +242,14 @@ where
 /// it refrains from passing any data to it.
 fn try_with_pin<F>(
   ctx: &mut args::ExecCtx<'_>,
-  pin_type: pinentry::PinType,
+  pin_entry: &pinentry::PinEntry,
   msg: &'static str,
   op: F,
 ) -> Result<()>
 where
   F: Fn(&str) -> result::Result<(), nitrokey::CommandError>,
 {
-  try_with_pin_and_data(ctx, pin_type, msg, (), |data, pin| {
+  try_with_pin_and_data(ctx, pin_entry, msg, (), |data, pin| {
     op(pin).map_err(|err| (data, err))
   })
 }
@@ -296,12 +294,15 @@ pub fn status(ctx: &mut args::ExecCtx<'_>) -> Result<()> {
 /// Open the encrypted volume on the nitrokey.
 pub fn storage_open(ctx: &mut args::ExecCtx<'_>) -> Result<()> {
   let device = get_storage_device(ctx)?;
-  try_with_pin(
-    ctx,
-    pinentry::PinType::User,
-    "Opening encrypted volume failed",
-    |pin| device.enable_encrypted_volume(&pin),
-  )
+  let pin_entry = pinentry::PinEntry::from(pinentry::PinType::User, &device)?;
+
+  // We may forcefully close a hidden volume, if active, so be sure to
+  // flush caches to disk.
+  unsafe { sync() };
+
+  try_with_pin(ctx, &pin_entry, "Opening encrypted volume failed", |pin| {
+    device.enable_encrypted_volume(&pin)
+  })
 }
 
 /// Close the previously opened encrypted volume.
@@ -317,6 +318,60 @@ pub fn storage_close(ctx: &mut args::ExecCtx<'_>) -> Result<()> {
     .map_err(|err| get_error("Closing encrypted volume failed", err))
 }
 
+/// Create a hidden volume.
+pub fn storage_hidden_create(
+  ctx: &mut args::ExecCtx<'_>,
+  slot: u8,
+  start: u8,
+  end: u8,
+) -> Result<()> {
+  let device = get_storage_device(ctx)?;
+  let pwd_entry = pinentry::PwdEntry::from(&device)?;
+  let pwd = if let Some(pwd) = &ctx.password {
+    pwd
+      .to_str()
+      .ok_or_else(|| Error::from("Failed to read password: invalid Unicode data found"))
+      .map(ToOwned::to_owned)
+  } else {
+    pinentry::choose(&pwd_entry)
+  }?;
+
+  device
+    .create_hidden_volume(slot, start, end, &pwd)
+    .map_err(|err| get_error("Creating hidden volume failed", err))
+}
+
+/// Open a hidden volume.
+pub fn storage_hidden_open(ctx: &mut args::ExecCtx<'_>) -> Result<()> {
+  let device = get_storage_device(ctx)?;
+  let pwd_entry = pinentry::PwdEntry::from(&device)?;
+  let pwd = if let Some(pwd) = &ctx.password {
+    pwd
+      .to_str()
+      .ok_or_else(|| Error::from("Failed to read password: invalid Unicode data found"))
+      .map(ToOwned::to_owned)
+  } else {
+    pinentry::inquire(&pwd_entry, pinentry::Mode::Query, None)
+  }?;
+
+  // We may forcefully close an encrypted volume, if active, so be sure
+  // to flush caches to disk.
+  unsafe { sync() };
+
+  device
+    .enable_hidden_volume(&pwd)
+    .map_err(|err| get_error("Opening hidden volume failed", err))
+}
+
+/// Close a previously opened hidden volume.
+pub fn storage_hidden_close(ctx: &mut args::ExecCtx<'_>) -> Result<()> {
+  unsafe { sync() };
+
+  get_storage_device(ctx)?
+    .disable_hidden_volume()
+    .map_err(|err| get_error("Closing hidden volume failed", err))
+}
+
 /// Pretty print the status of a Nitrokey Storage.
 fn print_storage_status(
   ctx: &mut args::ExecCtx<'_>,
@@ -430,11 +485,7 @@ fn get_otp<T: GenerateOtp>(slot: u8, algorithm: args::OtpAlgorithm, device: &T)
 fn get_unix_timestamp() -> Result<u64> {
   time::SystemTime::now()
     .duration_since(time::UNIX_EPOCH)
-    .or_else(|_| {
-      Err(Error::Error(
-        "Current system time is before the Unix epoch".to_string(),
-      ))
-    })
+    .or_else(|_| Err(Error::from("Current system time is before the Unix epoch")))
     .map(|duration| duration.as_secs())
 }
 
@@ -488,8 +539,8 @@ fn prepare_ascii_secret(secret: &str) -> Result<String> {
   if secret.is_ascii() {
     Ok(format_bytes(&secret.as_bytes()))
   } else {
-    Err(Error::Error(
-      "The given secret is not an ASCII string despite --format ascii being set".to_string(),
+    Err(Error::from(
+      "The given secret is not an ASCII string despite --format ascii being set",
     ))
   }
 }
@@ -498,7 +549,7 @@ fn prepare_ascii_secret(secret: &str) -> Result<String> {
 fn prepare_base32_secret(secret: &str) -> Result<String> {
   base32::decode(base32::Alphabet::RFC4648 { padding: false }, secret)
     .map(|vec| format_bytes(&vec))
-    .ok_or_else(|| Error::Error("Could not parse base32 secret".to_string()))
+    .ok_or_else(|| Error::from("Could not parse base32 secret"))
 }
 
 /// Configure a one-time password slot on the Nitrokey device.
@@ -557,9 +608,7 @@ fn print_otp_status(
     slot = match slot.checked_add(1) {
       Some(slot) => slot,
       None => {
-        return Err(Error::Error(
-          "Integer overflow when iterating OTP slots".to_string(),
-        ))
+        return Err(Error::from("Integer overflow when iterating OTP slots"));
       }
     };
     let name = match result {
@@ -588,53 +637,27 @@ pub fn otp_status(ctx: &mut args::ExecCtx<'_>, all: bool) -> Result<()> {
 }
 
 /// Clear the PIN stored by various operations.
-pub fn pin_clear() -> Result<()> {
-  pinentry::clear_pin(pinentry::PinType::Admin)?;
-  pinentry::clear_pin(pinentry::PinType::User)?;
+pub fn pin_clear(ctx: &mut args::ExecCtx<'_>) -> Result<()> {
+  let device = get_device(ctx)?;
+
+  pinentry::clear(&pinentry::PinEntry::from(
+    pinentry::PinType::Admin,
+    &device,
+  )?)?;
+  pinentry::clear(&pinentry::PinEntry::from(pinentry::PinType::User, &device)?)?;
   Ok(())
 }
 
-fn check_pin(pin_type: pinentry::PinType, pin: &str) -> Result<()> {
-  let minimum_length = match pin_type {
-    pinentry::PinType::Admin => 8,
-    pinentry::PinType::User => 6,
-  };
-  if pin.len() < minimum_length {
-    Err(Error::Error(format!(
-      "The PIN must be at least {} characters long",
-      minimum_length
-    )))
-  } else {
-    Ok(())
-  }
-}
-
-fn choose_pin_with_pinentry(pin_type: pinentry::PinType) -> Result<String> {
-  pinentry::clear_pin(pin_type)?;
-  let new_pin = pinentry::inquire_pin(pin_type, pinentry::Mode::Choose, None)?;
-  pinentry::clear_pin(pin_type)?;
-  check_pin(pin_type, &new_pin)?;
-
-  let confirm_pin = pinentry::inquire_pin(pin_type, pinentry::Mode::Confirm, None)?;
-  pinentry::clear_pin(pin_type)?;
-
-  if new_pin != confirm_pin {
-    Err(Error::Error("Entered PINs do not match".to_string()))
-  } else {
-    Ok(new_pin)
-  }
-}
-
 /// Choose a PIN of the given type.
 ///
 /// If the user has set the respective environment variable for the
 /// given PIN type, it will be used.
 fn choose_pin(
   ctx: &mut args::ExecCtx<'_>,
-  pin_type: pinentry::PinType,
+  pin_entry: &pinentry::PinEntry,
   new: bool,
 ) -> Result<String> {
-  let new_pin = match pin_type {
+  let new_pin = match pin_entry.pin_type() {
     pinentry::PinType::Admin => {
       if new {
         &ctx.new_admin_pin
@@ -654,20 +677,22 @@ fn choose_pin(
   if let Some(new_pin) = new_pin {
     new_pin
       .to_str()
-      .ok_or_else(|| Error::Error("Failed to read PIN: invalid Unicode data found".into()))
+      .ok_or_else(|| Error::from("Failed to read PIN: invalid Unicode data found"))
       .map(ToOwned::to_owned)
   } else {
-    choose_pin_with_pinentry(pin_type)
+    pinentry::choose(pin_entry)
   }
 }
 
 /// Change a PIN.
 pub fn pin_set(ctx: &mut args::ExecCtx<'_>, pin_type: pinentry::PinType) -> Result<()> {
   let device = get_device(ctx)?;
-  let new_pin = choose_pin(ctx, pin_type, true)?;
+  let pin_entry = pinentry::PinEntry::from(pin_type, &device)?;
+  let new_pin = choose_pin(ctx, &pin_entry, true)?;
+
   try_with_pin(
     ctx,
-    pin_type,
+    &pin_entry,
     "Could not change the PIN",
     |current_pin| match pin_type {
       pinentry::PinType::Admin => device.change_admin_pin(&current_pin, &new_pin),
@@ -679,10 +704,13 @@ pub fn pin_set(ctx: &mut args::ExecCtx<'_>, pin_type: pinentry::PinType) -> Resu
 /// Unblock and reset the user PIN.
 pub fn pin_unblock(ctx: &mut args::ExecCtx<'_>) -> Result<()> {
   let device = get_device(ctx)?;
-  let user_pin = choose_pin(ctx, pinentry::PinType::User, false)?;
+  let pin_entry = pinentry::PinEntry::from(pinentry::PinType::User, &device)?;
+  let user_pin = choose_pin(ctx, &pin_entry, false)?;
+  let pin_entry = pinentry::PinEntry::from(pinentry::PinType::Admin, &device)?;
+
   try_with_pin(
     ctx,
-    pinentry::PinType::Admin,
+    &pin_entry,
     "Could not unblock the user PIN",
     |admin_pin| device.unlock_user_pin(&admin_pin, &user_pin),
   )
@@ -758,7 +786,7 @@ fn print_pws_slot(
   programmed: bool,
 ) -> Result<()> {
   if slot > u8::MAX as usize {
-    return Err(Error::Error("Invalid PWS slot number".to_string()));
+    return Err(Error::from("Invalid PWS slot number"));
   }
   let slot = slot as u8;
   let name = if programmed {
